"""Tests for AttrDict.from_str and load_config keyword sources."""

import pytest

from pi_conf import load_config
from pi_conf.attr_dict import AttrDict


def test_from_str_toml():
    cfg = AttrDict.from_str("foo = 1\n[bar]\na = 2", "toml")
    assert cfg.foo == 1
    assert cfg.bar.a == 2


def test_from_str_json():
    cfg = AttrDict.from_str('{"x": {"y": 3}}', "json")
    assert cfg.x.y == 3


def test_from_str_ini():
    cfg = AttrDict.from_str(
        "[section]\nkey = value\n",
        "ini",
    )
    assert cfg.section.key == "value"


def test_from_str_yaml():
    pytest.importorskip("yaml")
    cfg = AttrDict.from_str("a:\n  b: 4\n", "yaml")
    assert cfg.a.b == 4


def test_from_str_unknown_type():
    with pytest.raises(Exception, match="Unknown config_type"):
        AttrDict.from_str("{}", "xml")


def test_load_config_data_keyword():
    cfg = load_config(data={"a": 1, "b": {"c": 2}})
    assert cfg.a == 1
    assert cfg.b.c == 2


def test_load_config_path_keyword(tmp_path):
    p = tmp_path / "cfg.toml"
    p.write_text("k = 9\n")
    cfg = load_config(path=str(p))
    assert cfg.k == 9


def test_load_config_path_keyword_ignore_warnings(tmp_path):
    missing = tmp_path / "nope.toml"
    cfg = load_config(path=str(missing), ignore_warnings=True)
    assert dict(cfg) == {}


def test_load_config_appname_keyword(tmp_path, monkeypatch):
    monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
    monkeypatch.setenv("HOME", str(tmp_path))
    appdir = tmp_path / ".config" / "kwapp"
    appdir.mkdir(parents=True)
    (appdir / "config.toml").write_text("z = 7\n")
    cfg = load_config(appname="kwapp")
    assert cfg.z == 7


def test_load_config_appname_prefers_dot_config_toml(tmp_path, monkeypatch):
    monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
    monkeypatch.setenv("HOME", str(tmp_path))
    appdir = tmp_path / ".config" / "kwapp"
    appdir.mkdir(parents=True)
    (appdir / ".config.toml").write_text("z = 9\n")
    (appdir / "config.toml").write_text("z = 7\n")
    cfg = load_config(appname="kwapp")
    assert cfg.z == 9


def test_load_config_keyword_conflict_raises():
    with pytest.raises(ValueError, match="Only one of data="):
        load_config(data={"a": 1}, path="x.toml")


def test_load_config_positional_plus_keyword_raises(tmp_path):
    p = tmp_path / "c.toml"
    p.write_text("a = 1\n")
    with pytest.raises(ValueError, match="Cannot combine a positional"):
        load_config(str(p), path=str(p))


def test_load_config_file_param_with_appname_keyword(tmp_path, monkeypatch):
    monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
    monkeypatch.setenv("HOME", str(tmp_path))
    appdir = tmp_path / ".config" / "myapp"
    appdir.mkdir(parents=True)
    (appdir / "secrets.toml").write_text("token = \"abc\"\n")
    cfg = load_config(appname="myapp", file="secrets.toml")
    assert cfg.token == "abc"



import os
import tempfile
from pathlib import Path

import pytest

from pi_conf import load_config
from pi_conf.config import _load_config_file


def test_config_loads_toml():
    s = """
    [a]
    b = 1"""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".toml") as f:
        f.write(s)
        f.flush()
        cfg = _load_config_file(f.name)
        assert cfg["a"]["b"] == 1


def test_config_loads_json():
    s = """
    {
        "a": {
            "b": 1
        }
    }
    """
    with tempfile.NamedTemporaryFile(mode="w", suffix=".json") as f:
        f.write(s)
        f.flush()
        cfg = _load_config_file(f.name)
        assert cfg["a"]["b"] == 1


def test_config_loads_yaml():
    s = """
    a:
        b: 1
    """
    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as f:
        f.write(s)
        f.flush()
        cfg = _load_config_file(f.name)
        assert cfg["a"]["b"] == 1


def test_config_loads_ini():
    s = """
    [a]
    b = 1
    """
    with tempfile.NamedTemporaryFile(mode="w", suffix=".ini") as f:
        f.write(s)
        f.flush()
        cfg = _load_config_file(f.name)
        assert cfg["a"]["b"] == "1"


def test_load_from_directory(tmpdir):
    s = """
    [a]
    b = 1
    """
    with open(os.path.join(tmpdir, "alt_config.toml"), "w") as f:
        f.write(s)
    cfg = load_config("alt_config.toml", directories=[tmpdir])

    assert cfg["a"]["b"] == 1


def test_load_non_default_nested(tmpdir):
    s = """
    [a]
    b = 1
    """
    nested_dir = os.path.join(tmpdir, "nested")
    os.makedirs(nested_dir)
    file_name = f"alt_config.toml"
    file_loc = os.path.join(tmpdir, os.path.join(nested_dir, file_name))
    with open(file_loc, "w") as f:
        f.write(s)
    cfg = load_config(file_name, directories=[nested_dir])

    assert cfg["a"]["b"] == 1


def test_load_non_default_nested_filename(tmpdir):
    s = """
    [a]
    b = 1
    """
    nested_dir = os.path.join(tmpdir, "nested")
    os.makedirs(nested_dir)
    file_name = f"alt_config.toml"
    file_loc = os.path.join(tmpdir, os.path.join(nested_dir, file_name))
    with open(file_loc, "w") as f:
        f.write(s)
    cfg = load_config(f"nested/{file_name}", directories=[tmpdir])

    assert cfg["a"]["b"] == 1


def test_load_from_abs_path(tmpdir):
    s = """
    [a]
    b = 1
    """
    file_name = f"alt_config.toml"
    file_loc = os.path.join(tmpdir, file_name)
    with open(file_loc, "w") as f:
        f.write(s)
    cfg = load_config(file_loc)

    assert cfg["a"]["b"] == 1


def test_config_loads_toml_from_path():
    s = """
    [a]
    b = 1"""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".toml", delete=False) as f:
        f.write(s)
        f.flush()
        path = Path(f.name)

    cfg = _load_config_file(path)
    assert cfg["a"]["b"] == 1

    path.unlink()  # Clean up the temporary file


def test_load_from_directory_with_path(tmpdir):
    s = """
    [a]
    b = 1
    """
    config_path = Path(tmpdir) / "alt_config.toml"
    config_path.write_text(s)

    cfg = load_config("alt_config.toml", directories=[Path(tmpdir)])
    assert cfg["a"]["b"] == 1


def test_load_non_default_nested_with_path(tmpdir):
    s = """
    [a]
    b = 1
    """
    nested_dir = Path(tmpdir) / "nested"
    nested_dir.mkdir(parents=True, exist_ok=True)

    file_name = "alt_config.toml"
    file_loc = nested_dir / file_name
    file_loc.write_text(s)

    cfg = load_config(file_name, directories=[nested_dir])
    assert cfg["a"]["b"] == 1


def test_load_from_abs_path_with_path(tmpdir):
    s = """
    [a]
    b = 1
    """
    file_name = "alt_config.toml"
    file_loc = Path(tmpdir) / file_name
    file_loc.write_text(s)

    cfg = load_config(file_loc)
    assert cfg["a"]["b"] == 1


def test_load_config_with_different_file_types_using_path(tmpdir):
    base_path = Path(tmpdir)

    # TOML
    toml_path = base_path / "config.toml"
    toml_path.write_text("[a]\nb = 1")

    # JSON
    json_path = base_path / "config.json"
    json_path.write_text('{"a": {"b": 1}}')

    # YAML
    yaml_path = base_path / "config.yaml"
    yaml_path.write_text("a:\n  b: 1")

    # INI
    ini_path = base_path / "config.ini"
    ini_path.write_text("[a]\nb = 1")

    for path in [toml_path, json_path, yaml_path, ini_path]:
        cfg = _load_config_file(path)
        assert cfg["a"]["b"] == 1 if path.suffix != ".ini" else "1"


if __name__ == "__main__":
    pytest.main([__file__])



""" Custom BaseSettings class for loading in complex types from toml files 
using the Config class."""

import json
import re
import tempfile
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Self, Type, TypeVar, get_args, get_origin

try:
    import toml as toml_io
except ImportError:
    toml_io = None

from pydantic import BaseModel, PrivateAttr, ValidationError
from pydantic_settings import BaseSettings, SettingsConfigDict

from pi_conf import Config, load_config
from pi_conf.config import load_from_appname

# Check if pymongo is installed
try:
    from pymongo import MongoClient

    MONGODB_AVAILABLE = True
except ImportError:
    MONGODB_AVAILABLE = False

sentinel = object()
M = TypeVar('M', bound=BaseModel)

class ConfigDict(SettingsConfigDict, total=False):
    """Extended SettingsConfigDict for TOML-specific settings.

    Use ``config_file`` for an explicit path to a config file. Do not use pydantic-settings'
    ``toml_file`` key here: that name is reserved for pydantic's built-in TOML source and
    triggers a UserWarning if set without that source.

    Attributes:
        config_file: Path to a config file (``.toml``, ``.json``, etc.).
        appname (str): The name of the application.
        toml_table_header (str): Header for the TOML table.
    """

    config_file: str
    appname: str
    toml_table_header: str
    mongo_uri: str
    mongo_database: str
    mongo_collection: str
    mongo_query: dict[str, Any]


class ConfigSource(ABC):
    @abstractmethod
    def load_config(self) -> Config: ...

    @abstractmethod
    def update_config(self, updates: Dict[str, Any]) -> None: ...

    @abstractmethod
    def refresh_config(self) -> Config: ...


@dataclass
class TomlConfigSource(ConfigSource):
    toml_file: Optional[Path]
    appname: Optional[str]
    toml_table_header: str = ""

    def load_config(self) -> Config:
        if self.toml_file and self.appname:
            cfg = load_from_appname(appname=self.appname, file=self.toml_file)
        elif self.toml_file:
            cfg = load_config(self.toml_file)
        elif self.appname:
            cfg = load_config(self.appname)
        else:
            raise ValueError("Either toml_file or appname must be provided")

        if self.toml_table_header:
            try:
                cfg = cfg.get_nested(self.toml_table_header)
            except KeyError:
                raise KeyError(
                    f"toml_table_header '{self.toml_table_header}' not found in config {cfg.provenance[-1].source}"
                )

        return cfg

    def update_config(self, updates: dict[str, Any]) -> None:
        if not self.toml_file:
            raise ValueError("Cannot update config: No TOML file specified")
        if toml_io is None:
            raise ImportError(
                "Writing TOML requires the 'toml' package (e.g. pip install toml). "
                "It is included when installing pi-conf with the pydantic extra."
            )

        current_config = toml_io.load(self.toml_file)

        if self.toml_table_header:
            table = current_config
            for key in self.toml_table_header.split("."):
                table = table.setdefault(key, {})
            table.update(updates)
        else:
            current_config.update(updates)

        with open(self.toml_file, "w") as f:
            toml_io.dump(current_config, f)

    def refresh_config(self) -> Config:
        return self.load_config()


if MONGODB_AVAILABLE:
    from bson import ObjectId

    @dataclass
    class MongoConfigSource(ConfigSource):
        mongo_uri: str
        mongo_database: str
        mongo_collection: str
        mongo_query: dict[str, Any]
        _id: ObjectId = field(init=False)

        def _extract_hostname(self, uri: str) -> str:
            match = re.search(r"@([^/:]+)", uri)
            return match.group(1) if match else "unknown-host"

        def load_config(self) -> Config:
            client: MongoClient = MongoClient(self.mongo_uri)
            try:
                db = client[self.mongo_database]
                collection = db[self.mongo_collection]
                config = collection.find_one(self.mongo_query)
                if config is None:
                    hostname = self._extract_hostname(self.mongo_uri)
                    raise ValueError(
                        f"No configuration found in mongodb+srv://{hostname} "
                        f"{self.mongo_database}.{self.mongo_collection} query: {self.mongo_query}"
                    )
                if config:
                    self._id = ObjectId(config.pop("_id"))
                return Config(config)
            finally:
                client.close()

        def update_config(self, updates: dict[str, Any]) -> None:
            if self._id is None:
                raise ValueError("No document ID available. Make sure to load the config first.")

            client: MongoClient = MongoClient(self.mongo_uri)
            try:
                db = client[self.mongo_database]
                collection = db[self.mongo_collection]

                update_result = collection.update_one({"_id": self._id}, {"$set": updates})

                if update_result.matched_count == 0:
                    raise ValueError(f"No document found with ID {self._id}")
            finally:
                client.close()

        def refresh_config(self) -> Config:
            return self.load_config()

        def insert_config(self, insert: dict[str, Any]) -> ObjectId:
            client: MongoClient = MongoClient(self.mongo_uri)
            try:
                db = client[self.mongo_database]
                collection = db[self.mongo_collection]
                insert_result = collection.insert_one(insert)
                self._id = insert_result.inserted_id
                return self._id
                # return Config.from_dict(self.mongo_query)
            finally:
                client.close()


class ConfigSettings(BaseSettings):
    """Main class for handling configuration settings.

    Attributes:
        model_config (ConfigDict): Configuration for the model.
    """

    model_config = ConfigDict(toml_table_header="")
    _config_source: ConfigSource = PrivateAttr(default=None)

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        raw = cls.__dict__.get("model_config")
        if not isinstance(raw, dict) or "toml_file" not in raw:
            return
        migrated = {k: v for k, v in raw.items() if k != "toml_file"}
        legacy_path = raw.get("toml_file")
        if legacy_path not in (None, ""):
            migrated.setdefault("config_file", legacy_path)
        cls.model_config = ConfigDict(**migrated)

    def __init__(self, *args, **kwargs):
        source_kw = (
            "config_file",
            "toml_file",
            "appname",
            "toml_table_header",
            "mongo_uri",
            "mongo_database",
            "mongo_collection",
            "model_config",
        )
        if kwargs and not any(var in kwargs for var in source_kw):
            return super().__init__(*args, **kwargs)

        base_mc = kwargs.pop("model_config", self.model_config)
        model_config = dict(base_mc)
        model_config.pop("toml_file", None)

        cf = kwargs.pop("config_file", sentinel)
        tf = kwargs.pop("toml_file", sentinel)
        if cf is not sentinel:
            model_config["config_file"] = cf
        elif tf is not sentinel:
            model_config["config_file"] = tf

        specified_vars = [
            "appname",
            "toml_table_header",
            "mongo_uri",
            "mongo_database",
            "mongo_collection",
        ]
        for var in specified_vars:
            val = kwargs.pop(var, sentinel)
            if val is not sentinel:
                model_config[var] = val

        _config_source = self._get_config_source(model_config)
        try:
            cfg = _config_source.load_config()
        except Exception as e:
            ## Add class name to error message
            e.args = (f"{self.__class__.__name__}: {e.args[0] if e.args else ''}",) + e.args[1:]
            raise
        try:
            self._parse_nested_objects(cfg)
            super().__init__(*args, **cfg, **kwargs)
            self._config_source = _config_source
        finally:
            pass

    @classmethod
    def _from_toml(
        cls,
        toml_file: Optional[Path] = None,
        appname: Optional[str] = None,
        toml_table_header: str = "",
    ):
        config_source = TomlConfigSource(toml_file, appname, toml_table_header)
        return cls(config_source)

    @classmethod
    def _from_mongodb(
        cls, mongo_uri: str, mongo_database: str, mongo_collection: str, mongo_query: dict[str, Any]
    ):
        if not MONGODB_AVAILABLE:
            raise ImportError(
                "MongoDB support requires pymongo to be installed. Please install it with 'pip install pymongo'."
            )
        config_source = MongoConfigSource(mongo_uri, mongo_database, mongo_collection, mongo_query)
        return cls(config_source)

    @staticmethod
    def from_config(model_class: Type[M], **kwargs) -> M:
        # Create a temporary class that inherits from both ConfigSettings and the model_class
        class TempConfigModel(ConfigSettings, model_class):  # type: ignore
            pass

        # Use this temporary class to load the configuration
        config_settings = TempConfigModel(**kwargs)

        # Extract only the fields defined in the original model_class
        model_fields = {
            field: getattr(config_settings, field) for field in model_class.model_fields
        }

        # Validate and return an instance of the original model_class
        return model_class.model_validate(model_fields)

    def _update(self, updates: dict[str, Any]) -> None:
        if self._config_source is None:
            raise ValueError(
                "ConfigSettings was constructed without a TOML/Mongo source; _update is not available."
            )
        self._config_source.update_config(updates)
        self.__dict__.update(updates)

    def _refresh(self) -> None:
        if self._config_source is None:
            raise ValueError(
                "ConfigSettings was constructed without a TOML/Mongo source; _refresh is not available."
            )
        new_config = self._config_source.refresh_config()
        self.__dict__.update(new_config)

    def _get_config_source(self, model_config: ConfigDict) -> ConfigSource:
        if "mongo_uri" in model_config:
            if MONGODB_AVAILABLE:
                for key in ["mongo_uri", "mongo_database", "mongo_collection"]:
                    if key not in model_config:
                        raise ValueError(f"MongoDB configuration requires {key} to be provided")
                mongo_uri = model_config["mongo_uri"]
                mongo_database = model_config["mongo_database"]  # type: ignore
                mongo_collection = model_config["mongo_collection"]  # type: ignore
                mongo_query = model_config["mongo_query"]  # type: ignore

                return MongoConfigSource(mongo_uri, mongo_database, mongo_collection, mongo_query)
            else:
                raise ImportError(
                    "MongoDB support requires pymongo to be installed. Please install it with 'pip install pymongo'."
                )
        elif (
            model_config.get("config_file")
            or model_config.get("toml_file")
            or "appname" in model_config
        ):
            config_path = model_config.get("config_file") or model_config.get("toml_file")
            if config_path:
                if isinstance(config_path, (list, tuple)):
                    config_path = Path(*config_path).expanduser()
                else:
                    config_path = Path(config_path).expanduser()
            appname = model_config.get("appname")
            toml_table_header = model_config.get("toml_table_header", "")

            if not (config_path or appname):
                raise ValueError("Either TOML file or appname must be provided")
            assert config_path is None or isinstance(config_path, Path)
            return TomlConfigSource(
                toml_file=config_path,
                appname=appname,
                toml_table_header=toml_table_header,
            )
        else:
            raise ValueError("Either TOML or MongoDB configuration must be provided")

    @classmethod
    def model_construct(cls, _fields_set: set[str] | None = None, **values: Any) -> Self:
        """Construct the model.

        Args:
            _fields_set: Set of fields to be set (not currently supported).
            **values: Arbitrary keyword arguments for model construction.

        Returns:
            Self: An instance of the class.

        Raises:
            NotImplementedError: If _fields_set is provided.
        """
        if _fields_set is not None:
            raise NotImplementedError(
                "ConfigSettings.model_construct does not currently support _fields_set"
            )

        return cls(**values)

    @classmethod
    def _create_temp_toml_file(cls, config: Config) -> str:
        """Create a temporary TOML file from the given configuration.

        Args:
            config (Config): The configuration object.

        Returns:
            str: The path to the created temporary TOML file.
        """
        if toml_io is None:
            raise ImportError(
                "Writing TOML requires the 'toml' package (e.g. pip install toml)."
            )
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".toml", delete=False) as temp_file:
            toml_io.dump(config, temp_file)
            return temp_file.name

    def _parse_nested_objects(self, config_dict: Dict[str, Any]):
        """Parse nested objects in the configuration dictionary.

        Args:
            config_dict (Dict[str, Any]): The configuration dictionary to parse.
        """
        for field_name, field_value in self.__annotations__.items():
            if field_name in config_dict:
                config_dict[field_name] = self._parse_field(field_value, config_dict[field_name])

    def _parse_field(self, field_type, field_value):
        """Parse a single field based on its type.

        Args:
            field_type: The type of the field.
            field_value: The value of the field.

        Returns:
            The parsed field value.
        """
        origin = get_origin(field_type)
        if origin is List:
            item_type = get_args(field_type)[0]
            if isinstance(item_type, type) and issubclass(item_type, BaseModel):
                return [self._parse_model(item_type, item) for item in field_value]
        elif isinstance(field_type, type) and issubclass(field_type, BaseModel):
            return self._parse_model(field_type, field_value)
        return field_value

    def _parse_model(self, model_class: Type[BaseModel], data: Dict[str, Any]) -> BaseModel:
        """Parse and validate a model using Pydantic.

        Args:
            model_class (Type[BaseModel]): The Pydantic model class.
            data (Dict[str, Any]): The data to parse into the model.

        Returns:
            BaseModel: An instance of the parsed model.

        Raises:
            ValidationError: If the data fails to validate against the model.
        """
        try:
            return model_class(**data)
        except ValidationError as e:
            print(f"Error parsing {model_class.__name__}: {e}")
            raise e

    def _pformat(self, indent: int = 4) -> str:
        return json.dumps(self.model_dump(mode="json"), indent=indent)



"""
ConfigSettings: subclasses pydantic_settings.BaseSettings.

Use this when you want typed fields, validation, nested BaseModel sections, and
the usual settings machinery (e.g. environment variables can still apply for
fields not filled from TOML, per pydantic-settings behavior).

Install: pip install 'pi-conf[pydantic]' (or your package manager equivalent).
"""

from __future__ import annotations

import sys
from pathlib import Path

_ROOT = Path(__file__).resolve().parents[1]
if str(_ROOT) not in sys.path:
    sys.path.insert(0, str(_ROOT))

from pydantic import BaseModel

from pi_conf.config_settings import ConfigDict, ConfigSettings


class Database(BaseModel):
    host: str
    timeout_seconds: int


class DemoSettings(ConfigSettings):
    app_name: str
    port: int
    database: Database

    model_config = ConfigDict(toml_table_header="")


def main() -> None:
    path = Path(__file__).resolve().parent / "sample_config.toml"
    settings = DemoSettings(config_file=path)

    print("ConfigSettings (Pydantic BaseSettings + TOML)")
    print(f"  app_name: {settings.app_name!r}")
    print(f"  port: {settings.port}")
    print(f"  database: {settings.database.model_dump()}")


if __name__ == "__main__":
    main()



import os
import sys

import pytest

import pi_conf.config as config_module
from pi_conf import AttrDict, Config, load_config, set_config

basedir = os.path.abspath(os.getcwd())
sys.path.append(basedir)


def test_config_empty_at_start():
    from pi_conf import cfg

    assert len(cfg) == 0


def test_config_from_dict():
    cfg = Config.from_dict({"a": 1})
    assert cfg.a == 1


def test_config_from_raw_dict():
    cfg = Config({"a": 1})
    assert cfg.a == 1


def test_config_from_list_dict_nested():
    cfg = Config.from_dict({"a": 1, "b": [{"c": 2}, {"d": 3}]})
    assert cfg.a == 1
    assert cfg.b[0].c == 2


def test_config_dict_init():
    cfg = Config({"a": 1})
    assert cfg.a == 1
    assert cfg["a"] == 1
    assert cfg.get("a") == 1


def test_to_env_list():
    cfg = Config({"a": [{"b": 1}, {"b": 2}]})
    envs = cfg.to_env(overwrite=True)
    assert envs == [("A0_B", "1"), ("A1_B", "2")]


def test_to_env_list_dict():
    cfg = Config({"a": [{"b": 1}, {"b": {"c": 2}}]})
    envs = cfg.to_env(overwrite=True)
    assert envs == [("A0_B", "1"), ("A1_B_C", "2")]


def test_to_env_dict():
    cfg = Config({"a": {"b": 1, "c": {"d": 2}}})
    envs = cfg.to_env(overwrite=True)
    assert envs == [("A_B", "1"), ("A_C_D", "2")]


def test_to_env_str():
    cfg = Config({"a": 1, "b": "2"})
    envs = cfg.to_env(overwrite=True)
    assert envs == [("A", "1"), ("B", "2")]


def test_set_config_not_exists(tmpdir):
    bn = os.path.basename(tmpdir)
    set_config(bn, directories=[tmpdir])
    with open(os.path.join(tmpdir, "config.toml"), "w") as f:
        f.write("a = 1")
        f.flush()
    cfg = load_config(bn, directories=[tmpdir])
    assert cfg.a == 1


def test_update_attrdict_with_attrdict():
    d1 = AttrDict({"a": 1, "b": 2})
    d2 = AttrDict({"a": 2, "c": 3})
    d1.update(d2)
    assert d1.a == 2
    assert d1.b == 2
    assert d1.c == 3


def test_update_attrdict_with_dict():
    d1 = AttrDict({"a": 1, "b": 2})
    d2 = {"b": {"a": 2, "c": 3}}
    d1.update(d2)
    assert d1.a == 1
    assert d1.b.a == 2
    assert d1.b.c == 3


def test_config_load_config_conflict_raises_by_default():
    cfg = Config.from_dict({"a": 1})

    with pytest.raises(ValueError, match="overwrite existing keys: a"):
        cfg.load_config({"a": 2})


def test_config_load_config_overwrite_true_allows_replacement():
    cfg = Config.from_dict({"a": 1})

    cfg.load_config({"a": 2}, overwrite=True)

    assert cfg.a == 2


def test_update_config_conflict_raises_by_default():
    original = Config.from_dict(dict(config_module.cfg))
    original_provenance = list(config_module.cfg.provenance)
    try:
        config_module.cfg.clear()
        config_module.cfg.update({"a": 1}, _add_to_provenance=False)

        with pytest.raises(ValueError, match="overwrite existing keys: a"):
            config_module.update_config({"a": 2}, directories=None)
    finally:
        config_module.cfg.clear()
        config_module.cfg.update(original, _add_to_provenance=False)
        pm = config_module.get_pmanager()
        pm.clear(config_module.cfg)
        pm.extend(config_module.cfg, original_provenance)


def test_update_config_overwrite_true_allows_replacement():
    original = Config.from_dict(dict(config_module.cfg))
    original_provenance = list(config_module.cfg.provenance)
    try:
        config_module.cfg.clear()
        config_module.cfg.update({"a": 1}, _add_to_provenance=False)

        updated = config_module.update_config({"a": 2}, directories=None, overwrite=True)

        assert updated.a == 2
        assert config_module.cfg.a == 2
    finally:
        config_module.cfg.clear()
        config_module.cfg.update(original, _add_to_provenance=False)
        pm = config_module.get_pmanager()
        pm.clear(config_module.cfg)
        pm.extend(config_module.cfg, original_provenance)


if __name__ == "__main__":
    pytest.main([__file__])



from unittest.mock import Mock, patch

import pytest
from bson import ObjectId
from pydantic import BaseModel
from pymongo.errors import ServerSelectionTimeoutError

from pi_conf.config_settings import ConfigDict, ConfigSettings, MongoConfigSource


class SubConfig(BaseModel):
    value: int


class MyConfig(ConfigSettings):
    string_value: str
    int_value: int
    nested_config: SubConfig

    model_config = ConfigDict()


class MockCursor:
    def __init__(self, documents):
        self.documents = documents

    def __iter__(self):
        return iter(self.documents)


class MockCollection:
    def __init__(self, documents):
        self.documents = documents

    def find_one(self, query):
        return self.documents[0] if self.documents else None

    def find(self, query):
        return MockCursor(self.documents)


class MockDatabase:
    def __init__(self, collections):
        self.collections = collections

    def __getitem__(self, name):
        return self.collections.get(name, MockCollection([]))


class MockMongoClient:
    def __init__(self, databases):
        self.databases = databases

    def __getitem__(self, name):
        return self.databases.get(name, MockDatabase({}))

    def close(self):
        pass


@pytest.fixture
def mock_mongo():
    test_document = {
        "_id": ObjectId(),
        "string_value": "test_string",
        "int_value": 42,
        "nested_config": {"value": 10},
    }
    return MockMongoClient(
        {"test_db": MockDatabase({"test_collection": MockCollection([test_document])})}
    )


@patch("pi_conf.config_settings.MongoClient")
def test_mongo_config_source(MockClient, mock_mongo):
    MockClient.return_value = mock_mongo

    config_source = MongoConfigSource(
        mongo_uri="mongodb://localhost:27017",
        mongo_database="test_db",
        mongo_collection="test_collection",
        mongo_query={"string_value": "test_string"},
    )
    config = config_source.load_config()

    assert isinstance(config, dict)
    assert config["string_value"] == "test_string"
    assert config["int_value"] == 42
    assert config["nested_config"]["value"] == 10


@patch("pi_conf.config_settings.MongoClient")
def test_config_settings_with_mongo(MockClient, mock_mongo):
    MockClient.return_value = mock_mongo

    mongo_uri = "mongodb://localhost:27017"
    db_name = "test_db"
    collection_name = "test_collection"

    class MongoTestConfig(MyConfig):
        model_config = ConfigDict(
            mongo_uri=mongo_uri, mongo_database=db_name, mongo_collection=collection_name
        )

    config = MongoTestConfig(
        string_value="default_string", int_value=42, nested_config={"value": 10}  # type: ignore
    )

    assert config.string_value == "default_string"
    assert config.int_value == 42
    assert isinstance(config.nested_config, SubConfig)
    assert config.nested_config.value == 10


@patch("pi_conf.config_settings.MongoClient")
def test_mongo_config_source_empty_collection(MockClient):
    mock_mongo = MockMongoClient({"test_db": MockDatabase({"test_collection": MockCollection([])})})
    MockClient.return_value = mock_mongo

    config_source = MongoConfigSource(
        mongo_uri="mongodb://localhost:27017",
        mongo_database="test_db",
        mongo_collection="test_collection",
        mongo_query={"string_value": "test_string"},
    )

    with pytest.raises(ValueError, match="No configuration found in mongodb"):
        config_source.load_config()

@patch("pi_conf.config_settings.MongoClient")
def test_mongo_config_source_multiple_documents(MockClient):
    mock_mongo = MockMongoClient(
        {
            "test_db": MockDatabase(
                {
                    "test_collection": MockCollection(
                        [
                            {"_id": ObjectId(), "doc1": "value1"},
                            {"_id": ObjectId(), "doc2": "value2"}
                        ]
                    )
                }
            )
        }
    )
    MockClient.return_value = mock_mongo

    config_source = MongoConfigSource(
        mongo_uri="mongodb://localhost:27017",
        mongo_database="test_db",
        mongo_collection="test_collection",
        mongo_query={"string_value": "test_string"},
    )
    config = config_source.load_config()

    # Assuming MongoConfigSource returns the first document when multiple are present
    assert config == {"doc1": "value1"}


@patch("pi_conf.config_settings.MongoClient")
def test_mongo_connection_error(MockClient):
    MockClient.side_effect = ServerSelectionTimeoutError("Connection error")

    config_source = MongoConfigSource(
        mongo_uri="mongodb://localhost:27017",
        mongo_database="test_db",
        mongo_collection="test_collection",
        mongo_query={"string_value": "test_string"},
    )

    with pytest.raises(ServerSelectionTimeoutError, match="Connection error"):
        config_source.load_config()


if __name__ == "__main__":
    pytest.main([__file__])



import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional

import pytest
import toml
from pydantic import BaseModel, ValidationError

from pi_conf.config_settings import ConfigDict, ConfigSettings


@pytest.fixture
def test_data():
    return {
        "string_value": "test",
        "int_value": 42,
        "list_value": [1, 2, 3],
        "dict_value": {"a": 1, "b": 2},
        "pymodel_value": {"name": "Test", "sub": {"value": 10}},
        "pymodel_list": [
            {"name": "Test1", "sub": {"value": 1}},
            {"name": "Test2", "sub": {"value": 2}},
        ],
    }


class SubModel(BaseModel):
    value: int


class ModelValue(BaseModel):
    name: str
    sub: SubModel


class MySettings(ConfigSettings):
    string_value: Optional[str] = None
    int_value: Optional[int] = None
    list_value: Optional[List[int]] = None
    dict_value: Optional[Dict[str, int]] = None
    pymodel_value: Optional[ModelValue] = None
    pymodel_list: Optional[List[ModelValue]] = None

    model_config = ConfigDict()


def config_settings(toml_file_path: str) -> MySettings:
    return MySettings.model_construct(
        model_config=ConfigDict(appname=toml_file_path),
    )


def write_toml(data: dict[str, Any], path: Path, name: str = "config.toml"):
    toml_string = toml.dumps(data)
    with open(path / name, "w") as toml_file:
        toml_file.write(toml_string)


def test_config_file_kwarg_no_toml_file_source_warning(tmp_path, test_data):
    """Regression: avoid pydantic-settings UserWarning for unused built-in toml_file."""
    write_toml(test_data, tmp_path)
    path = tmp_path / "config.toml"
    with warnings.catch_warnings(record=True) as caught:
        warnings.simplefilter("always")
        MySettings(config_file=str(path))
    bad = [
        w
        for w in caught
        if issubclass(w.category, UserWarning) and "TomlConfigSettingsSource" in str(w.message)
    ]
    assert not bad


def test_legacy_model_config_toml_file_migrated(tmp_path):
    partial_data = {"string_value": "legacy", "int_value": 7}
    write_toml(partial_data, tmp_path)
    p = str(tmp_path / "config.toml")

    class LegacySettings(ConfigSettings):
        string_value: Optional[str] = None
        int_value: Optional[int] = None
        model_config = ConfigDict(toml_file=p)

    assert "toml_file" not in LegacySettings.model_config
    assert LegacySettings.model_config.get("config_file") == p
    s = LegacySettings()
    assert s.string_value == "legacy"
    assert s.int_value == 7


def test_config_settings(tmp_path, test_data):
    config_file = tmp_path / "config.toml"
    write_toml(test_data, tmp_path)
    settings = config_settings(str(config_file))
    assert settings.string_value == test_data["string_value"]
    assert settings.int_value == test_data["int_value"]
    assert settings.list_value == test_data["list_value"]
    assert settings.dict_value == test_data["dict_value"]
    assert settings.pymodel_value
    assert settings.pymodel_value.model_dump() == test_data["pymodel_value"]
    assert settings.pymodel_list
    assert [model.model_dump() for model in settings.pymodel_list] == test_data["pymodel_list"]


def test_missing_optional_values(tmp_path):
    partial_data = {
        "string_value": "test",
        "int_value": 42,
    }
    write_toml(partial_data, tmp_path)
    settings = config_settings(str(tmp_path / "config.toml"))
    assert settings.string_value == "test"
    assert settings.int_value == 42
    assert settings.list_value is None
    assert settings.dict_value is None
    assert settings.pymodel_value is None
    assert settings.pymodel_list is None


def test_invalid_data_type(tmp_path):
    invalid_data = {
        "int_value": "not an integer",
    }
    write_toml(invalid_data, tmp_path)
    with pytest.raises(ValidationError):
        config_settings(str(tmp_path / "config.toml"))


def test_nested_toml_table(tmp_path):
    nested_data = {
        "some_other_section": {},
        "my_section": {
            "string_value": "nested",
            "int_value": 100,
        },
    }
    write_toml(nested_data, tmp_path)

    class NestedSettings(ConfigSettings):
        string_value: str
        int_value: int
        model_config = ConfigDict(toml_table_header="my_section")

    settings = NestedSettings.model_construct(
        model_config=ConfigDict(
            appname=str(tmp_path / "config.toml"), toml_table_header="my_section"
        ),
    )
    assert settings.string_value == "nested"
    assert settings.int_value == 100


def test_config_dict_construction():
    config = ConfigDict(appname="test_app", toml_table_header="my_section")
    assert "appname" in config and "toml_table_header" in config
    assert config["appname"] == "test_app"
    assert config["toml_table_header"] == "my_section"


def test_model_construct_with_values():
    values = {
        "string_value": "constructed",
        "int_value": 200,
    }
    settings = MySettings.model_construct(**values)
    assert settings.string_value == "constructed"
    assert settings.int_value == 200
    assert settings.list_value is None


@pytest.mark.parametrize("file_name", ["config.toml", "alternate_config.toml"])
def test_different_file_names(tmp_path, test_data, file_name):
    write_toml(test_data, tmp_path, name=file_name)
    settings = config_settings(str(tmp_path / file_name))
    assert settings.string_value == test_data["string_value"]
    assert settings.int_value == test_data["int_value"]


def test_from_config(tmp_path):
    # Define a simple Pydantic model for testing
    class UserPreferences(BaseModel):
        theme: str
        notifications_enabled: bool
        favorite_numbers: List[int]
        display_name: Optional[str] = None

    # Create test data
    test_data = {
        "theme": "dark",
        "notifications_enabled": True,
        "favorite_numbers": [42, 7, 13],
        "display_name": "TestUser",
    }

    # Write test data to TOML file
    write_toml(test_data, tmp_path)

    # Use from_config to create the model instance
    user_prefs = ConfigSettings.from_config(
        UserPreferences, model_config={"appname": str(tmp_path / "config.toml")}
    )

    # Verify the model was created correctly
    assert isinstance(user_prefs, UserPreferences)
    assert user_prefs.theme == "dark"
    assert user_prefs.notifications_enabled is True
    assert user_prefs.favorite_numbers == [42, 7, 13]
    assert user_prefs.display_name == "TestUser"


def test_from_config_with_nested_model(tmp_path):
    # Define nested models
    class Address(BaseModel):
        street: str
        city: str
        postal_code: str

    class User(BaseModel):
        name: str
        age: int
        address: Address
        tags: Optional[List[str]] = None

    # Create test data with nested structure
    test_data = {
        "name": "John Doe",
        "age": 30,
        "address": {"street": "123 Main St", "city": "Springfield", "postal_code": "12345"},
        "tags": ["user", "premium"],
    }

    # Write test data to TOML file
    write_toml(test_data, tmp_path)

    # Use from_config to create the model instance
    user = ConfigSettings.from_config(User, model_config={"appname": str(tmp_path / "config.toml")})

    # Verify the model and its nested components
    assert isinstance(user, User)
    assert user.name == "John Doe"
    assert user.age == 30
    assert isinstance(user.address, Address)
    assert user.address.street == "123 Main St"
    assert user.address.city == "Springfield"
    assert user.address.postal_code == "12345"
    assert user.tags == ["user", "premium"]


def test_from_config_with_missing_optional(tmp_path):
    # Define a model with optional fields
    class Settings(BaseModel):
        required_field: str
        optional_field: Optional[str] = None
        optional_list: Optional[List[str]] = None

    # Create test data with only required field
    test_data = {"required_field": "value"}

    # Write test data to TOML file
    write_toml(test_data, tmp_path)

    # Use from_config to create the model instance
    settings = ConfigSettings.from_config(
        Settings, model_config={"appname": str(tmp_path / "config.toml")}
    )

    # Verify the model handles missing optional fields
    assert isinstance(settings, Settings)
    assert settings.required_field == "value"
    assert settings.optional_field is None
    assert settings.optional_list is None


def test_from_config_validation_error(tmp_path):
    # Define a model with specific validation
    from pydantic import BaseModel, Field, ValidationError

    class ValidatedSettings(BaseModel):
        port: int = Field(ge=1, le=65535)  # Using Field for validation

    # Create test data with invalid port
    test_data = {"port": 70000}  # Invalid port number

    # Write test data to TOML file
    write_toml(test_data, tmp_path)

    # Verify that validation error is raised
    with pytest.raises(ValidationError):
        ConfigSettings.from_config(
            ValidatedSettings, model_config={"appname": str(tmp_path / "config.toml")}
        )


# Alternative version using model validator
def test_from_config_validation_error_with_validator(tmp_path):
    from pydantic import BaseModel, ValidationError, field_validator

    class ValidatedSettings(BaseModel):
        port: int

        @field_validator("port")
        @classmethod
        def validate_port(cls, v: int) -> int:
            if not 1 <= v <= 65535:
                raise ValueError("Port must be between 1 and 65535")
            return v

    # Create test data with invalid port
    test_data = {"port": 70000}  # Invalid port number

    # Write test data to TOML file
    write_toml(test_data, tmp_path)

    # Verify that validation error is raised
    with pytest.raises(ValidationError):
        ConfigSettings.from_config(
            ValidatedSettings, model_config={"appname": str(tmp_path / "config.toml")}
        )


class PlainSettings(ConfigSettings):
    model_config = ConfigDict()
    count: int = 0


def test_config_settings_update_without_file_source():
    s = PlainSettings(count=1)
    with pytest.raises(ValueError, match="without a TOML/Mongo source"):
        s._update({"count": 2})
    with pytest.raises(ValueError, match="without a TOML/Mongo source"):
        s._refresh()


if __name__ == "__main__":
    pytest.main([__file__])



import os
import sys
from dataclasses import dataclass, field

import pytest
from pi_conf import AttrDict, Config

basedir = os.path.abspath(os.getcwd())
sys.path.append(basedir)

@pytest.fixture
def myconfig():
    @dataclass
    class MyConfig(Config):
        a: int = 1
    return MyConfig

def test_dataclass_subclass(myconfig):
    cfg = myconfig()
    assert cfg.a == 1

def test_dataclass_subclass_from_dict(myconfig):
    cfg = myconfig.from_dict({"a": 1})
    assert cfg.a == 1
    assert type(cfg) == myconfig

def test_config_dict_init():
    @dataclass
    class MyConfig(Config):
        a: int = 1
        b: int = 2

    cfg = MyConfig(**{"a": 1, "b": 2})
    assert cfg.a == 1
    assert cfg["a"] == 1
    assert cfg.get("a") == 1
    assert cfg.b == 2
    assert cfg["b"] == 2
    assert cfg.get("b") == 2

def test_dataclass_subclass_from_dict_recursive_nested_false():
    @dataclass
    class MyConfig(Config):
        a: int = 1
        b: AttrDict = field(default_factory=AttrDict)

    cfg = MyConfig.from_dict({"a": 1, "b": {"c": 2}})
    assert cfg.a == 1
    assert cfg["a"] == 1
    assert cfg.b.c == 2
    assert cfg["b"]["c"] == 2
    assert type(cfg) == MyConfig
    assert type(cfg.a) == int
    assert type(cfg.b) == AttrDict


if __name__ == "__main__":
    pytest.main([__file__])


"""
Config: a provenance-aware dict (not Pydantic).

Use this when you want flexible nested data from TOML/JSON/YAML, attribute
access like cfg.database.host, and provenance for where values were loaded from.
This is not a substitute for pydantic-settings: there are no declared fields,
environment-variable overrides, or automatic coercion beyond what the file parser does.
"""

from __future__ import annotations

import sys
from pathlib import Path

_ROOT = Path(__file__).resolve().parents[1]
if str(_ROOT) not in sys.path:
    sys.path.insert(0, str(_ROOT))

from pi_conf import load_config


def main() -> None:
    path = Path(__file__).resolve().parent / "sample_config.toml"
    cfg = load_config(path=str(path))

    print("Config (dict-like + attributes)")
    print(f"  app_name: {cfg.app_name!r}")
    print(f"  port: {cfg.port}")
    print(f"  database.host: {cfg.database.host!r}")
    print(f"  provenance entries: {len(cfg.provenance)}")


if __name__ == "__main__":
    main()



"""
ConfigSettings from a nested TOML table (not the file root).

Set ``toml_table_header`` to the dot path of the table (e.g. ``"worker"`` or
``"services.api"``). Pi-conf loads the file, then slices that subtree before
building the Pydantic model, so your fields match only that section.

This uses ``sample_layered_config.toml``, which also contains app-level keys;
those are ignored when loading ``WorkerSettings`` because only the ``[worker]``
slice is passed to Pydantic.

Install: pip install 'pi-conf[pydantic]' (or your package manager equivalent).
"""

from __future__ import annotations

import sys
from pathlib import Path

_ROOT = Path(__file__).resolve().parents[1]
if str(_ROOT) not in sys.path:
    sys.path.insert(0, str(_ROOT))

from pi_conf.config_settings import ConfigDict, ConfigSettings


class WorkerSettings(ConfigSettings):
    """Settings that live under ``[worker]`` in sample_layered_config.toml."""

    queue_name: str
    concurrency: int

    model_config = ConfigDict(toml_table_header="worker")


def main() -> None:
    path = Path(__file__).resolve().parent / "sample_layered_config.toml"
    settings = WorkerSettings(config_file=path)

    print("ConfigSettings scoped to [worker] (toml_table_header)")
    print(f"  queue_name: {settings.queue_name!r}")
    print(f"  concurrency: {settings.concurrency}")


if __name__ == "__main__":
    main()



import pytest

from pi_conf import AttrDict


@pytest.fixture
def sample_dict():
    return AttrDict({"a": {"b": {"c": 1, "d": None}, "e": 2}, "f": 3, "g": None})


def test_get_nested_existing_keys(sample_dict):
    assert sample_dict.get_nested("a.b.c") == 1
    assert sample_dict.get_nested("a.e") == 2
    assert sample_dict.get_nested("f") == 3
    assert sample_dict.get_nested("a.b.d") is None
    assert sample_dict.get_nested("g") is None


def test_get_nested_missing_keys_with_default(sample_dict):
    assert sample_dict.get_nested("a.b.x", "default") == "default"
    assert sample_dict.get_nested("x.y.z", None) is None
    assert sample_dict.get_nested("a.x", 0) == 0


def test_get_nested_missing_keys_without_default(sample_dict):
    with pytest.raises(KeyError):
        sample_dict.get_nested("a.b.x")
    with pytest.raises(KeyError):
        sample_dict.get_nested("x.y.z")


def test_get_nested_non_dict_intermediate(sample_dict):
    with pytest.raises(KeyError):
        sample_dict.get_nested("f.x")
    assert sample_dict.get_nested("f.x", "default") == "default"


def test_get_nested_empty_key(sample_dict):
    with pytest.raises(KeyError):
        sample_dict.get_nested("")


def test_get_nested_none_default(sample_dict):
    assert sample_dict.get_nested("a.b.x", None) is None
    assert sample_dict.get_nested("x.y.z", None) is None


def test_get_nested_modify_dict(sample_dict):
    sample_dict["h"] = {"i": {"j": 4}}
    assert sample_dict.get_nested("h.i.j") == 4


def test_get_nested_with_special_characters(sample_dict):
    sample_dict["special!@#"] = {"$%^&*": {"()_+": "chars"}}
    assert sample_dict.get_nested("special!@#.$%^&*.()_+") == "chars"


def test_get_nested_list_behavior():
    d = AttrDict(
        {
            "a": [{"b": 1}, {"c": 2}],
            "x": [{"y": [{"z": 3}]}],
        },
    )

    # Test basic list behavior
    assert d.get_nested("a") == [{"b": 1}, {"c": 2}]

    # Test accessing first list item by default
    assert d.get_nested("a.b") == 1

    # Test accessing second list item
    assert d.get_nested("a.c", list_item=1) == 2

    # Test deeper nesting
    assert d.get_nested("x.y.z") == 3

    # Test deeper nesting with default
    assert d.get_nested("x.y.z.notexist", default=4) == 4

    # Test error when key doesn't exist in first list item
    with pytest.raises(KeyError):
        d.get_nested("a.c")

    # Test using default value
    assert d.get_nested("a.d", default=None) is None


def test_get_nested_list_item_parameter():
    d = AttrDict({"a": [{"b": 1}, {"b": 2}, {"b": 3}]})

    # Test different list_item values
    assert d.get_nested("a.b", list_item=0) == 1
    assert d.get_nested("a.b", list_item=1) == 2
    assert d.get_nested("a.b", list_item=2) == 3

    # Test out of range
    with pytest.raises(KeyError):
        d.get_nested("a.b", list_item=3)

    # Test with default value
    assert d.get_nested("a.b", list_item=3, default="default") == "default"


def test_get_nested_disable_list_item():
    d = AttrDict({"a": [{"b": 1}, {"b": 2}]})

    # Test disabling list_item
    assert d.get_nested("a", list_item=None) == [{"b": 1}, {"b": 2}]


def test_get_nested_list_item():
    d = AttrDict({"a": [{"b": 1}, {"b": 2}]})

    assert d.get_nested("a") == [{"b": 1}, {"b": 2}]


if __name__ == "__main__":
    pytest.main(["-v", __file__])



import importlib.util


def check_module(module_name: str) -> bool:
    return importlib.util.find_spec(module_name) is not None


has_yaml = check_module("yaml")
has_stdlib_tomllib = check_module("tomllib")
has_toml_package = check_module("toml")



import os
import sys
import tempfile
import pytest

import pi_conf.config as config
from pi_conf import Config
from pi_conf.provenance import ProvenanceOp

basedir = os.path.abspath(os.getcwd())
sys.path.append(basedir)

def test_provenance_of_config_provenance_disabled():
    from pi_conf.provenance import _provenance_manager

    cfg = Config({"a": 1}, enable_provenance=False)

    assert len(cfg.provenance) == 0
    oid = id(cfg)
    assert oid not in _provenance_manager._enabled

def test_provenance_from_dict_directly():
    cfg = Config({"a": 1})
    assert cfg.provenance[-1].source == "dict"
    assert cfg.provenance[-1].operation == ProvenanceOp.set
    assert len(cfg.provenance) == 1

def test_provenance_from_dict():
    cfg = Config.from_dict({"a": 1})
    assert cfg.provenance[-1].source == "dict"
    assert cfg.provenance[-1].operation == ProvenanceOp.set
    assert len(cfg.provenance) == 1

def test_provenance_from_dict2():
    cfg = Config.from_dict({"a": 1})
    assert cfg.provenance[-1].source == "dict"
    assert cfg.provenance[-1].operation == ProvenanceOp.set
    cfg.update({"b": 2})
    assert cfg.provenance[-1].source == "dict"
    assert cfg.provenance[-1].operation == ProvenanceOp.update
    assert len(cfg.provenance) == 2

def test_provenance_from_set_config():
    s = """
    [a]
    b = 1"""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".toml") as f:
        f.write(s)
        f.flush()

        cfg = config.set_config(f.name)

        assert len(cfg.provenance) == 1
        assert cfg.provenance[0].source == f.name
        assert cfg.provenance[0].operation == ProvenanceOp.set

def test_provenance_from_load_config():
    s = """
    [a]
    b = 1"""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".toml") as f:
        f.write(s)
        f.flush()

        cfg = config.load_config(f.name)

        assert len(cfg.provenance) == 1
        assert cfg.provenance[0].source == f.name

def test_provenance_clear_clears_instance_not_global():
    global_cfg = config.cfg
    len_global_before = len(global_cfg.provenance)
    other = Config.from_dict({"x": 1})
    assert len(other.provenance) >= 1
    other.clear()
    assert len(other.provenance) == 0
    assert len(global_cfg.provenance) == len_global_before


def test_provenance_from_load_config_twice():
    s = """
    [a]
    b = 1"""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".toml") as f:
        f.write(s)
        f.flush()

        cfg = config.load_config(f.name)
        cfg = config.load_config(f.name)

        assert len(cfg.provenance) == 1
        assert cfg.provenance[0].source == f.name
        pm = config.get_pmanager()
        pm._provenance

if __name__ == "__main__":
    pytest.main([__file__])